﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Perforce.P4
{
    class P4ServerMT : IDisposable
    {
        private String server;
        private String user;
        private String pass;
        private String ws_client;

        private Dictionary<int, P4Server> mapTIDtoServer = new Dictionary<int, P4Server>();
        private Object dictLock = new object();
        private string cwd;
        private string trustFlag;
        private string fingerprint;

        public P4ServerMT(String _server,
                         String _user,
                         String _pass,
                         String _ws_client)
        {
            server = _server;
            user = _user;
            pass = _pass;
            ws_client = _ws_client;
        }

        public P4ServerMT(string _server, string _user, string _pass, string _ws_client, string _cwd, string trustFlag, string fingerprint) : this(_server, _user, _pass, _ws_client)
        {
            this.cwd = _cwd;
            this.trustFlag = trustFlag;
            this.fingerprint = fingerprint;
        }

        public P4ServerMT(string _cwd)
        {
            this.cwd = _cwd;
        }

        public void Dispose()
        {
            // dispose of the P4Servers
            foreach (KeyValuePair<int, P4Server> entry in mapTIDtoServer)
            {
                entry.Value.Dispose();
            }
        }

        public P4Server getServer()
        {
            return getServerForThread(System.Threading.Thread.CurrentThread.ManagedThreadId);
        }

        public P4Server getServerForThread(int threadId)
        {
            lock (dictLock)
            {
                if (mapTIDtoServer.ContainsKey(threadId))
                {
                    return mapTIDtoServer[threadId];
                }
                // only call the fingerprint constructor if we got configured with a fingerprint
                // otherwise it will not throw the correct "you need to trust this" exception
                // (which seems wrong, both methods should operate similarly))
                P4Server p4server = (fingerprint != null && fingerprint.Length != 0) || 
                                    (trustFlag != null && trustFlag.Length != 0) ?
                    new P4Server(server, user, pass, ws_client, cwd, trustFlag, fingerprint) :
                    new P4Server(server, user, pass, ws_client, cwd);
                p4server.SetThreadOwner(threadId);
                mapTIDtoServer[threadId] = p4server;
                return p4server;
            }
        }


    }
}
